/* Copyright (C) 2025-2025 Warsaw University of Technology, Warsaw
    SPDX-License-Identifier: GPL-3.0-only
    Authors: Daniel Wielanek [committer] */
#include "HalCbmV0TempTrack.h"

#include <TLorentzVector.h>
#include <TVector3.h>

#include <Hal/Event.h>
#include <Hal/ExpTrack.h>
#include <Hal/Track.h>
#include <Hal/V0Track.h>

Double_t HalCbmV0TempTrack::GetValue(kTriggerCut val)
{
  switch (val) {
    case kTriggerCut::kChi2Geo: {
      return fV0Hidden.GetChi2geo();
    } break;
    case kTriggerCut::kChi2PrimFirst: {
      return fV0Hidden.GetChi2Pos();
    } break;
    case kTriggerCut::kChi2PrimSecond: {
      return fV0Hidden.GetChi2Neg();
    } break;
    case kTriggerCut::kChi2Topo: {
      //return fV0Hidden.GetChi2topo();
    } break;
    case kTriggerCut::kCosineFirst: {
      return fCosinePos;
    } break;
    case kTriggerCut::kCosineSecond: {
      return fCosineNeg;
    } break;
    case kTriggerCut::kDistance: {
      return fV0Hidden.GetChi2Pos();
    } break;
    case kTriggerCut::kL: {
      return fV0Hidden.GetDecLength();
    } break;
    case kTriggerCut::kDl: {
      return fV0Hidden.GetDecDl();
    } break;
    case kTriggerCut::kMass: {
      return 0;
    } break;
    case kTriggerCut::kPt: {
      return fRecoTrack.GetMomentum().Pt();
    } break;
    case kTriggerCut::kEta: {
      return fRecoTrack.GetMomentum().Eta();
    } break;
    case kTriggerCut::kRxy: {
      return fRecoTrack.GetDCA().Pt();
    } break;
    case kTriggerCut::kRz: {
      return fRecoTrack.GetDCA().Z();
    } break;
    case kTriggerCut::kCosineTopo: {
      return fCosineTopo;
    } break;
    case kTriggerCut::kPid: {
      return fV0Hidden.GetPdg();
    } break;
  }
  return 0;
}

void HalCbmV0TempTrack::ComputeStuff(Hal::Event* event, Hal::Track* assignedTrack)
{
  auto dau1     = (HalCbmTrack*) event->GetTrack(fV0Hidden.GetPosId());
  auto dau2     = (HalCbmTrack*) event->GetTrack(fV0Hidden.GetNegId());
  TVector3 vtx  = event->GetVertex()->Vect();
  auto expTrack = (Hal::ExpTrack*) assignedTrack;
  TVector3 pos(expTrack->GetDCA());  //or + event pos?
  TVector3 mom(expTrack->GetPx(), expTrack->GetPy(), expTrack->GetPz());
  auto shift = mom.Unit() * fV0Hidden.GetDecLength() + vtx;  //new position + pos gorsze wyniki
  fV0Hidden.SetDecayPos(shift.X(), shift.Y(), shift.Z());
  TVector3 dau1mom, dau2mom, dau1pos, dau2pos;
  dau1pos = dau1->GetHelix().Eval(shift.Z(), dau1mom);
  dau2pos = dau2->GetHelix().Eval(shift.Z(), dau2mom);
#ifdef DEBUG_EXTR
  std::cout << "DECPOS " << std::endl;
  std::cout << __FILE__ << "____" << std::endl;
  auto print = [](TString t, TVector3 x) {
    std::cout << Form("%4.4f %4.4f %4.4f %s", x.X(), x.Y(), x.Z(), t.Data()) << std::endl;
  };
  print("Dau1dec", dau1pos);
  print("Dau2dec", dau2pos);

  print("Dau1vec", dau1mom);
  print("Dau2vec", dau2mom);
#endif
  fV0Hidden.SetMomPos(dau1mom.X(), dau1mom.Y(), dau1mom.Z());
  fV0Hidden.SetMomNeg(dau2mom.X(), dau2mom.Y(), dau2mom.Z());
  fV0Hidden.Recalc(*assignedTrack);
}
