/* Copyright (C) 2023-2023 Warsaw University of Technology, Warsaw
    SPDX-License-Identifier: GPL-3.0-only
    Authors: Daniel Wielanek [committer] */
#include "CbmAnaTreeSource.h"

#include "CbmAnaTreeContainer.h"
#include "HalCbmDetectorID.h"

#include <FairLogger.h>
#include <FairRootManager.h>

#include <TBranch.h>
#include <TChain.h>
#include <TFile.h>

#include <AnalysisTree/Configuration.hpp>
#include <fstream>
#include <vector>

#include <Hal/InputDataInfo.h>


CbmAnaTreeSource::CbmAnaTreeSource() : CbmAnaTreeSource("", "aTree") {}

CbmAnaTreeSource::CbmAnaTreeSource(TString inFile, TString treeName)
  : fNFiles(1)
  , fTreeName(treeName)
  , fChain(nullptr)
  , fFileName(nullptr)
  , fATSource(nullptr)
{
  fFileName    = new TString[1];
  fFileName[0] = inFile;
}

CbmAnaTreeSource::~CbmAnaTreeSource()
{
  if (fFileName) delete[] fFileName;
}

Bool_t CbmAnaTreeSource::Init()
{
  LOG(debug) << "Init source ";
  FairRootManager* mngr = FairRootManager::Instance();
  fChain                = new TChain(fTreeName);  //new name of tree?

  Hal::InputDataInfo info;
  if (fFileName[0].EndsWith(".root")) {
    for (int i = 0; i < fNFiles; i++) {
      LOG(debug) << "CbmAnaTree: opening single file" << fFileName[i];
      fChain->Add(fFileName[i]);
      info.AddFile(fFileName[i]);
    }
  }
  else {  // this is long list
    std::ifstream list;
    list.open(fFileName[0]);
    bool first = true;
    do {
      TString temp;
      list >> temp;
      if (temp.Length() > 1) {
        fChain->Add(temp);
        info.AddFile(temp);
        if (first) {
          first = false;
        }
      }
      else {
        break;
      }
      LOG(debug) << "Adding file " << temp << " to chain";
    } while (!list.eof());
    list.close();
  }


  info.AddFile(mngr->GetInFile()->GetName());
  fATSource = new CbmAnaTreeContainers();
  LOG(debug) << "load container ";
  fATSource->LoadConf(&info);

  mngr->SetInChain(fChain, -1);

  auto contNames = CbmAnaTreeContainers::GetNames();
  for (int i = 0; i < fATSource->GetSize(); i++) {
    bool con = fATSource->fContainers[i]->ConnectToTree(fChain);
    if (con) {
      TString name = HalCbm::GetContainerName(contNames[i]);
      mngr->Register(name, "AT", fATSource->fContainers[i], kFALSE);
    }
  }
  return kTRUE;
}

Int_t CbmAnaTreeSource::ReadEvent(UInt_t unsignedInt)
{
  fChain->GetEntry(unsignedInt);
  return 0;
}

void CbmAnaTreeSource::Close() {}

void CbmAnaTreeSource::AddFile(TString file)
{
  TString* temp = fFileName;
  fFileName     = new TString[fNFiles + 1];
  for (int i = 0; i < fNFiles; i++) {
    fFileName[i] = temp[i];
  }
  delete[] temp;
  fFileName[fNFiles] = file;
  fNFiles            = fNFiles + 1;
}

Int_t CbmAnaTreeSource::CheckMaxEventNo(Int_t /*int1*/) { return fChain->GetEntries(); }
